--  << ui2.e >> -- graphical user interface system
--  Colin Taylor - cetaylor@compuserve.com
--      version 2.0   12/20/98

include graphics.e
include image.e
include mighty.e
include font.e

global constant INACTIVE = 0, NORMAL = 1, SET = 2
global sequence keys, key_data, buttons, bullets, things, zones  -- lists
global object mouse_event
global integer default_mouse_cursor, mbox

sequence sys_colors  -- {foreground, background, highlight, shadow}
sequence imgstack, winstack  -- saved images, saved lists
sequence shade  -- shade bitmap
integer key, found

keys = {}       -- key_val
key_data = {}   -- {status, proc_id}
buttons = {}    -- {p1, p2, status, label, proc_id}
bullets = {}    -- {p1, p2, status, p0, label, proc_id}
things = {}     -- {p1, p2, status, proc_id}
zones = {}      -- {p1, p2, status, inside, cursor, proc_id}
imgstack = {}
winstack = {}
default_mouse_cursor = IDC_ARROW
mbox = routine_id("graphics_mode") -- dispose of id=0, reserved for "exit"

global function tile(sequence bm, sequence a)
-- tiles bitmap bm over area a, where a = {width, height} in pixels.
    integer ax, ay, bmx, bmy
    sequence tbm
    ax = a[1]
    ay = a[2]
    bmx = length(bm[1])
    bmy = length(bm)
    if ax > bmx then
	tbm = repeat({}, bmy)
	for j = 1 to bmy do
	    for i = 1 to floor(ax/bmx)+1 do
		tbm[j] = tbm[j] & bm[j]
	    end for
	    tbm[j] = tbm[j][1..ax] --trim excess width
	end for
    else
	tbm = repeat({}, ay)
	for j = 1 to ay do
	    tbm[j] = bm[j][1..ax]
	end for
    end if
    if ay > bmy then
	bm = tbm
	for i = 1 to floor(ay/bmy)+1 do
	    tbm = tbm & bm
	end for
    end if
    return tbm[1..ay]  -- trim excess height
end function  -- tile

global procedure set_sys_colors(sequence colors)
-- sets ui and text colors
    sys_colors = colors
    set_foreground(sys_colors[1])   -- font.e colors
    set_background(sys_colors[2])   --  "
    set_shadow_color(sys_colors[4]) --  "
    shade = {{sys_colors[4], sys_colors[2]}, {sys_colors[2], sys_colors[4]}}
end procedure  -- set_sys_colors

global function get_sys_colors()
    return sys_colors
end function  -- get_sys_colors

global function mouse_box(sequence s)  -- inspired by J. Babor
-- scans a series of rectangles to see if mouse pointer is inside
    for i = 1 to length(s) do
	if s[i][3] then  -- check if object is active
	    if mouse_inside(s[i][1][1], s[i][1][2], s[i][2][1], s[i][2][2]) then
		mbox = i
		return mbox
	    end if
	end if
    end for
    return 0
end function  -- mouse_box

global procedure get_mbutton_release()
-- waits for mouse button release
    while mouse_event[1] do
	mouse_event = poll_mouse()
    end while
end procedure  -- get_mbutton_release

global procedure box(object c, integer f, integer os, sequence p1, sequence p2)
-- draws rectangle of color c using coordinates p1 and p2, enlarged by os
--    (flag f indicates outline (0) or filled (1))
    if sequence(c) then
	display_image(tile(c, {p1-os, p2+os}), p1-os)
    else
	polygon(c, f, {p1-os, {p2[1]+os, p1[2]-os}, p2+os, {p1[1]-os, p2[2]+os}})
    end if
end procedure  -- box

global procedure box_3d(integer z, integer os, sequence p1, sequence p2)
-- draws a raised or inset frame using coordinates p1, p2
--  z = height (-1 or 1), os = offset from corner coordinates
    integer c1, c2
    if z = -1 then
	c1 = sys_colors[4]
	c2 = sys_colors[3]
    else
	c1 = sys_colors[3]
	c2 = sys_colors[4]
    end if
    draw_line(c1, {{p1[1]-os, p2[2]+os-1}, p1-os, {p2[1]+os-1, p1[2]-os}})
    draw_line(c2, {{p1[1]-os+1, p2[2]+os}, p2+os, {p2[1]+os, p1[2]-os+1}})
end procedure  -- box_3d

global procedure text(sequence p, sequence label)
-- writes left-justified label at location p
    justify(LEFT)
    setx(p[1])
    sety(p[2])
    mouse_pointer(0)
    write(label)
    mouse_pointer(1)
    set_attributes(NONE)
    set_foreground(sys_colors[1])
end procedure  -- text

global procedure text_box(sequence p1, sequence p2, sequence label)
-- writes label centered in box {p1, p2}
    sequence len
    len = lengths(label)
    justify(CENTRE)
    setx(floor((p1[1]+p2[1])/2)+1)
    sety(floor((p1[2]+p2[2]-height()*length(len))/2)+1)
    mouse_pointer(0)
    write(label)
    mouse_pointer(1)
    justify(LEFT)
    set_attributes(NONE)
    set_foreground(sys_colors[1])
end procedure  -- text_box

global procedure panel(sequence p1, sequence p2)
-- draws a raised box filled with the background color
    box(0, 0, 0, p1, p2) -- always black
    box(sys_colors[2], 1, -1, p1, p2)
    box_3d(1, -1, p1, p2)
end procedure  -- panel

global procedure define_key(integer key_val, integer proc_id, integer status)
-- defines a key to be recognized by input_loop()
-- status:  INACTIVE or NORMAL
    integer found
    found = find(key_val, keys)
    if found then   -- redefine key
	key_data[found] = {status, proc_id}
    else            -- new key
	keys = append(keys, key_val)
	key_data = append(key_data, {status, proc_id})
    end if
end procedure  -- define_key

global procedure button(integer but_num, integer status)
-- draws a button with text label
    mouse_pointer(0)
    if status = INACTIVE then
	panel(buttons[but_num][1], buttons[but_num][2])
	set_foreground(shade)
	text_box(buttons[but_num][1], buttons[but_num][2], buttons[but_num][4])
    elsif status = NORMAL then
	panel(buttons[but_num][1], buttons[but_num][2])
	set_foreground(BLUE)
	text_box(buttons[but_num][1], buttons[but_num][2], buttons[but_num][4])
    else       -- status = SET
	box(0, 0, 0, buttons[but_num][1], buttons[but_num][2])
	display_image(buttons[but_num][1]+2, 
			tile(shade, buttons[but_num][2]-buttons[but_num][1]-3))
	box_3d(-1, -1, buttons[but_num][1], buttons[but_num][2])
	set_background(CLEAR)
	text_box(buttons[but_num][1], buttons[but_num][2], 
		    {271, -4} & buttons[but_num][4])
	set_background(sys_colors[2])
    end if
    mouse_pointer(1)
    buttons[but_num][3] = status
end procedure  -- button

global procedure button_press(integer but_num)
-- button press and release (using mouse-button)
    button(but_num, 2)
    get_mbutton_release()
    button(but_num, 1)
end procedure  -- button_press

global procedure define_button(sequence p1, sequence p2, sequence label, 
			integer proc_id, integer status)
-- adds a new button to the button list and draws button
-- status:  INACTIVE, NORMAL or SET
    buttons = append(buttons, {p1, p2, status, label, proc_id})
    button(length(buttons), status)
end procedure  -- define_button

global procedure bullet(integer bul_num, integer status)
-- draws a bullet
    mouse_pointer(0)
    if status = INACTIVE then
	box_3d(1, 5, bullets[bul_num][4], bullets[bul_num][4])
	display_image(bullets[bul_num][4]-3, tile(shade, {7, 7}))
    elsif status = NORMAL then
	box_3d(1, 5, bullets[bul_num][4], bullets[bul_num][4])
	box(sys_colors[2], 1, 3, bullets[bul_num][4], bullets[bul_num][4])
    else        -- SET
	box_3d(-1, 5, bullets[bul_num][4], bullets[bul_num][4])
	box(sys_colors[2], 1, 3, bullets[bul_num][4], bullets[bul_num][4])
	ellipse(sys_colors[1], 1, bullets[bul_num][4]-2, bullets[bul_num][4]+2)
    end if
    mouse_pointer(1)
    bullets[bul_num][3] = status
end procedure  -- bullet

global procedure define_bullet(sequence p0, sequence label, 
				integer proc_id, integer status)
-- creates a bullet and draws it together with label
-- status:  INACTIVE, NORMAL or SET
    sequence len, p1, p2
    integer ht
    ht = height()
    len = lengths(label)
    p1 = p0-6
    p2 = {p0[1]+len[1]+7, p0[2]}+6
    bullets = append(bullets, {p1, p2, status, p0, label, proc_id})
    bullet(length(bullets), status)
    text({p0[1]+12, p0[2]-floor(ht/2-1)}, label)
end procedure  -- define_bullet

global procedure thing(integer thi_num, integer status)
-- draws a graphic object
    mouse_pointer(0)
    if status = INACTIVE then
	display_image(things[thi_num][1], things[thi_num][4])
    elsif status = NORMAL then
	display_image(things[thi_num][1], things[thi_num][5])
    else         -- SET
	display_image(things[thi_num][1], things[thi_num][6])
    end if
    mouse_pointer(1)
    things[thi_num][3] = status
end procedure  -- thing

global procedure define_thing(sequence p1, sequence bm1, sequence bm2, 
			sequence bm3, integer proc_id, integer status)
-- create a graphic object and display it
-- bitmaps: bm1 = inactive image, bm2 = normal image, bm3 = set image
-- all bitmaps should be the same size
-- status:  INACTIVE, NORMAL or SET
    sequence p2
    p2 = p1+{length(bm2[1]), length(bm2)}-1
    things = append(things, {p1, p2, status, bm1, bm2, bm3, proc_id})
    thing(length(things), status)
end procedure  -- define_thing

global procedure cursor_zones()
-- set the mouse cursor according to predefined zones
    if mouse_box(zones) then
	if zones[mbox][4] = 0 then
	    zones[mbox][4] = 1  -- inside
	    set_mouse_cursor(zones[mbox][5])
	end if
    else
	for i = 1 to length(zones) do
	    if zones[i][4] then
		zones[i][4] = 0  -- outside
		set_mouse_cursor(default_mouse_cursor)
		exit
	    end if
	end for
    end if
end procedure  -- cursor_zones

global procedure define_zone(sequence p1, sequence p2, object cursor, 
			integer proc_id, integer status)
-- add a new cursor zone
-- status:  INACTIVE or NORMAL
    zones = append(zones, {p1, p2, status, 0, cursor, proc_id})
end procedure  -- define_zone

global procedure open_window(sequence p1, sequence p2)
-- draw panel after saving underlying screen image and window coords
-- windows can be stacked to any depth (memory permitting)
-- current clickable objects are pushed to winstack
    winstack = prepend(winstack, keys)
    keys = {}
    winstack = prepend(winstack, key_data)
    key_data = {}
    winstack = prepend(winstack, buttons)
    buttons = {}
    winstack = prepend(winstack, bullets)
    bullets = {}
    winstack = prepend(winstack, zones)
    zones = {}
    mouse_pointer(0)
    imgstack = prepend(imgstack, {p1, save_image(p1, p2)})
    panel(p1, p2)
    mouse_pointer(1)
end procedure  -- open_window

global procedure close_window()
-- closes the last window opened and restores the screen image
-- previous clickable objects are popped from winstack
    if length(imgstack) > 0 then
	mouse_pointer(0)
	display_image(imgstack[1][1], imgstack[1][2])
	mouse_pointer(1)
	imgstack = imgstack[2..length(imgstack)]
	zones = winstack[1]
	bullets = winstack[2]
	buttons = winstack[3]
	key_data = winstack[4]
	keys = winstack[5]
	winstack = winstack[6..length(winstack)]
    end if
end procedure  -- close_window

global procedure exit_loop()
-- pause for user action (key press or mouse click)
    while 1 do
	if get_key() != -1 then
	    exit
	end if
	mouse_event = poll_mouse()
	cursor_zones()
	if mouse_event[1] then
	    get_mbutton_release()
	    exit
	end if
    end while
end procedure  -- exit_loop

global procedure input_loop()
-- respond to keyboard or mouse input
-- turn mouse pointer on before calling input_loop()
    while 1 do
	-- key input scan
	key = get_key()
	if key = 27 then  -- escape always exits
	    exit
	else 
	    found = find(key, keys)
	    if found then 
		if key_data[found][1] then
		    if key_data[found][2] then
			call_proc(key_data[found][2], {})
		    else
			exit
		    end if
		end if
	    end if
	end if
	-- mouse input scan
	mouse_event = poll_mouse()
	cursor_zones()
	if mouse_event[1] = 1 then 
	    if mouse_box(buttons) then
		button_press(mbox)
		if buttons[mbox][5] then
		    call_proc(buttons[mbox][5], {})
		else
		    exit
		end if
	    elsif mouse_box(bullets) then
		get_mbutton_release()
		if bullets[mbox][6] then
		    call_proc(bullets[mbox][6], {})
		else
		    exit
		end if
	    elsif mouse_box(things) then
		get_mbutton_release()
		if things[mbox][7] then
		    call_proc(things[mbox][7], {})
		else
		    exit
		end if
	    elsif mouse_box(zones) then
		get_mbutton_release()
		if zones[mbox][6] then
		    call_proc(zones[mbox][6], {})
		else
		    exit
		end if
	    end if  
	end if
    end while
end procedure  -- input_loop

set_sys_colors({BLACK, WHITE, BRIGHT_WHITE, GRAY})  -- defaults

