-- << pal_util.e >> - palette utility routines
--  Colin Taylor  -  Jan 24, 1999

include machine.e
include image.e

constant EOF = -1
constant data = allocate(24), dither = allocate(487)
atom old_pal_addr, bm_addr, new_pal_addr, le_addr
atom bm_width, bm_height, new_pal_len, le_len
old_pal_addr = data
bm_addr = data+4
new_pal_addr = data+8
le_addr = data+12
bm_width = data+16
bm_height = data+18
new_pal_len = data+20
le_len = data+22

procedure poke2(atom addr, integer int)
    poke(addr, {remainder(int, #100), floor(int/#100)})
end procedure  -- poke2

global function color_usage(sequence image)
-- returns sequence showing usage of each palette color
    sequence pal, bm, index
    pal = image[1]
    bm = image[2]
    index = repeat(0, length(pal))
    for j = 1 to length(bm) do
	for i = 1 to length(bm[j]) do
	    index[bm[j][i]+1] = index[bm[j][i]+1]+1
	end for
    end for
    return index
end function  -- color_usage

global function unused_colors(sequence image)
-- returns a list of palette entries not used by image
    sequence index, list
    index = color_usage(image)
    list = {}
    for i = 1 to length(index) do
	if index[i] = 0 then
	    list = append(list, i-1)
	end if
    end for
    return list
end function  -- unused_colors

global function nearest_color(sequence pal, sequence rgb_color)
-- finds the closest color match to rgb_color in palette pal
    sequence err
    integer min_err, error, color_val
    min_err = 1073741823
    for i = 1 to length(pal) do
	err = rgb_color-pal[i]
	error = err[1]*err[1]+err[2]*err[2]+err[3]*err[3]
	if error = 0 then
	    return i-1
	elsif error < min_err then
	    min_err = error
	    color_val = i-1
	end if
    end for
    return color_val
end function  -- nearest_color

global function convert_image(sequence new_pal, sequence image)
-- remaps image to a new palette using the nearest colors in new_pal
    sequence old_pal, bitmap, index
    old_pal = image[1]
    bitmap = image[2]
    index = repeat(0, length(old_pal))
    for i = 1 to length(old_pal) do
	index[i] = nearest_color(new_pal, old_pal[i])
    end for
    for i = 1 to length(bitmap) do
	for j = 1 to length(bitmap[i]) do
	    bitmap[i][j] = index[bitmap[i][j]+1]
	end for
    end for
    return {new_pal, bitmap}
end function  -- convert_image

global function dither_image(sequence new_pal, sequence image)
-- remaps image to a new palette using colors from new_pal
-- dithers image to improve tone gradients (error diffusion)
    sequence old_pal, bm
    atom addr, npa, opa, bma, lea
    integer npl, opl, bmw, bmh, lel
    old_pal = image[1]
    bm = image[2]
    npl = length(new_pal)
    npa = allocate(npl*3)
    addr = npa
    for i = 1 to npl do
	poke(addr, new_pal[i])
	addr = addr+3
    end for
    opl = length(old_pal)
    opa = allocate(opl*3)
    addr = opa
    for i = 1 to opl do
	poke(addr, old_pal[i])
	addr = addr+3
    end for
    bmh = length(bm)
    bmw = length(bm[1])
    bma = allocate(bmw*bmh)
    addr = bma
    for i = 1 to bmh do
	poke(addr, bm[i])
	addr = addr+bmw
    end for
    lel = (bmw+2)*6
    lea = allocate(lel*2)
    poke4(old_pal_addr, opa)
    poke4(bm_addr, bma)
    poke2(bm_width, bmw)
    poke2(bm_height, bmh)
    poke4(new_pal_addr, npa)
    poke2(new_pal_len, npl)
    poke4(le_addr, lea)
    poke4(le_len, lel)
    call(dither)
    addr = bma
    for i = 1 to bmh do
	bm[i] = peek({addr, bmw})
	addr = addr+bmw
    end for
    free(npa)
    free(opa)
    free(bma)
    free(lea)
    return {new_pal, bm}
end function  -- dither_image

global function systemize_palette(sequence sys_pal, sequence old_pal)
-- reorganizes old_pal so that the colors matching sys_pal are 
-- placed in a block at the beginning of the palette
    sequence temp
    integer color
    if length(sys_pal)+length(old_pal) <= 256 then
	sys_pal = sys_pal & old_pal
	return sys_pal
    elsif length(old_pal) >= length(sys_pal) then
	for i = 1 to length(sys_pal) do
	    color = nearest_color(old_pal[i..length(old_pal)], sys_pal[i])+i-1
	    if color > i then
		temp = old_pal[i]
		old_pal[i] = old_pal[color+1]
		old_pal[color+1] = temp
	    end if
	end for
    return  old_pal
    else
	for i = 1 to length(old_pal) do
	    color = nearest_color(sys_pal, old_pal[i])
	    sys_pal[color+1] = old_pal[i]
	end for
	return sys_pal
    end if
end function  -- systemize_palette

global function systemize_image(sequence sys_pal, sequence image, integer exact)
-- systemizes palette and converts or dithers image to new palette
-- exact: 0 - use nearest colors and convert image to new palette
--        1 - use exact system colors and convert image to new palette
-- system colors will replace unused colors, if sufficient are available 
    sequence new_pal, unused
    if length(sys_pal)+length(image[1]) <= 256 then
	new_pal = sys_pal & image[1]
    else
	unused = unused_colors(image)
	if length(unused) > length(sys_pal) then
	    new_pal = image[1]
	    for i = 1 to length(sys_pal) do
		new_pal[unused[i]+1] = new_pal[i]
		new_pal[i] = sys_pal[i]
	    end for
	else
	    new_pal = systemize_palette(sys_pal, image[1])
	    if exact then
		new_pal[1..length(sys_pal)] = sys_pal
	    end if
	end if
    end if
    return convert_image(new_pal, image)
end function  -- systemize_image

global function save_palette(sequence pal, sequence name)
-- saves palette to a file in MS format, returns -1 for error
    integer fn, pl
    fn = open(name, "wb")
    pl = length(pal)
    if fn != -1 then
	puts(fn, "RIFF")
	puts(fn, int_to_bytes(pl*4+16))
	puts(fn, "PAL data")
	puts(fn, int_to_bytes(pl*4+4))
	puts(fn, {0, 3})  -- planes
	puts(fn, {remainder(pl, 256), floor(pl/256)})  -- colors
	for i = 1 to pl do
	    puts(fn, {pal[i][3], pal[i][2], pal[i][1], 0})
	end for
	close(fn)
    end if
    return fn
end function  -- save_palette

global function load_palette(sequence name)
-- loads palette from file in MS format, returns -1 for error
    sequence buffer, pal
    integer fn, pl, byte
    buffer = {}
    pal = {}
    fn = open(name, "rb")
    if fn != -1 then
	for i = 1 to 24 do
	    byte = getc(fn)
	    if byte = EOF then
		close(fn)
		return -1
	    end if
	    buffer = append(buffer, byte)
	end for
	if compare(buffer[1..4], "RIFF") then
	    close(fn)
	    return -1
	else
	    pl = buffer[23]+buffer[24]*256
	end if
	for i = 1 to pl do
	    buffer = {}
	    for j = 1 to 4 do
		byte = getc(fn)
		if byte = EOF then
		    close(fn)
		    return -1
		end if
		buffer = append(buffer, byte)
	    end for
	    pal = append(pal, {buffer[3], buffer[2], buffer[1]})
	end for
	close(fn)
	return pal
    end if
    return fn
end function  -- load_palette

-- dither converted to machine code by ASM.EX
poke(dither, {
#60,#89,#E5,#83,#EC,#0E,#FC,#B8,#00,#00,#00,#00,#8B,#0D,#00,#00,#00,#00,#01,#C9,
#8B,#3D,#00,#00,#00,#00,#F3,#AA,#B8,#06,#00,#00,#00,#89,#45,#F2,#8B,#3D,#00,#00,
#00,#00,#01,#C7,#89,#7D,#F6,#B9,#00,#00,#00,#00,#BA,#00,#00,#00,#00,#8B,#35,#00,
#00,#00,#00,#52,#51,#56,#8A,#1E,#B7,#00,#8B,#3D,#00,#00,#00,#00,#01,#DF,#01,#DF,
#01,#DF,#89,#EA,#83,#EA,#06,#8B,#75,#F6,#66,#B9,#00,#00,#8A,#07,#B4,#00,#66,#03,
#06,#66,#89,#02,#66,#FF,#C1,#83,#C2,#02,#83,#C6,#02,#47,#66,#83,#F9,#03,#7C,#E6,
#66,#BE,#FF,#7F,#8B,#3D,#00,#00,#00,#00,#66,#B9,#00,#00,#8A,#17,#B6,#00,#66,#2B,
#55,#FA,#79,#03,#66,#F7,#DA,#47,#8A,#07,#B4,#00,#66,#2B,#45,#FC,#79,#03,#66,#F7,
#D8,#66,#01,#C2,#47,#8A,#07,#B4,#00,#66,#2B,#45,#FE,#79,#03,#66,#F7,#D8,#66,#01,
#C2,#47,#66,#83,#FA,#60,#7F,#05,#66,#89,#CB,#EB,#17,#66,#39,#F2,#7D,#06,#66,#89,
#D6,#66,#89,#CB,#66,#FF,#C1,#66,#3B,#0D,#00,#00,#00,#00,#7C,#AE,#5E,#88,#1E,#8B,
#3D,#00,#00,#00,#00,#01,#DF,#01,#DF,#01,#DF,#66,#B9,#00,#00,#89,#EA,#83,#EA,#06,
#56,#8B,#75,#F6,#03,#75,#F2,#89,#75,#F6,#8A,#1F,#B7,#00,#66,#8B,#02,#66,#29,#D8,
#66,#83,#F8,#7F,#7E,#04,#66,#B8,#7F,#00,#66,#83,#F8,#81,#7D,#04,#66,#B8,#81,#FF,
#66,#D1,#F8,#66,#D1,#F8,#66,#89,#C3,#66,#D1,#F8,#66,#D1,#F8,#66,#01,#1E,#66,#01,
#1E,#66,#29,#06,#56,#03,#35,#00,#00,#00,#00,#66,#01,#06,#2B,#75,#F2,#66,#01,#1E,
#66,#01,#06,#2B,#75,#F2,#66,#01,#1E,#66,#29,#06,#66,#FF,#C1,#83,#C2,#02,#5E,#83,
#C6,#02,#47,#66,#83,#F9,#03,#0F,#8C,#99,#FF,#FF,#FF,#5E,#59,#5A,#66,#F7,#C2,#01,
#00,#75,#14,#46,#66,#FF,#C1,#66,#3B,#0D,#00,#00,#00,#00,#0F,#8C,#C3,#FE,#FF,#FF,
#4E,#EB,#0B,#4E,#66,#FF,#C9,#0F,#85,#B6,#FE,#FF,#FF,#46,#8B,#7D,#F6,#2B,#7D,#F2,
#89,#7D,#F6,#66,#FF,#C2,#66,#3B,#15,#00,#00,#00,#00,#7C,#05,#83,#C4,#0E,#61,#C3,
#8B,#45,#F2,#F7,#D8,#89,#45,#F2,#31,#C0,#66,#A1,#00,#00,#00,#00,#01,#C6,#56,#51,
#FC,#8B,#0D,#00,#00,#00,#00,#8B,#3D,#00,#00,#00,#00,#89,#FE,#01,#CE,#F3,#A4,#B8,
#00,#00,#00,#00,#8B,#0D,#00,#00,#00,#00,#8B,#3D,#00,#00,#00,#00,#01,#CF,#F3,#AA,
#59,#5E,#E9,#58,#FE,#FF,#FF})
poke4(dither + 14, le_len)
poke4(dither + 22, le_addr)
poke4(dither + 38, le_addr)
poke4(dither + 59, bm_addr)
poke4(dither + 72, old_pal_addr)
poke4(dither + 126, new_pal_addr)
poke4(dither + 210, new_pal_len)
poke4(dither + 221, new_pal_addr)
poke4(dither + 307, le_len)
poke4(dither + 370, bm_width)
poke4(dither + 409, bm_height)
poke4(dither + 432, bm_width)
poke4(dither + 443, le_len)
poke4(dither + 449, le_addr)
poke4(dither + 466, le_len)
poke4(dither + 472, le_addr)

