-- << pu_demo.ex >> -- palette management utilities demo
--  Colin Taylor - 1/24/99 - cetaylor@compuserve.com

include ui2.e
include pal_util.e
include cursors.e

sequence colors, max_size, vc, hex, image, palettes, file_name, file_err
integer saf, sbf, aus, upc, id, curr_pal, auto, load, dither, exact

max_size = {320, 240}
file_err = {"error opening file", "error reading file", 
	    "unsupported file format", "file error"}

procedure add_palette(object pal)
    if sequence(pal) then
	palettes = append(palettes, pal)
	bullet(length(palettes), NORMAL)
    else
	palettes = append(palettes, {})
    end if
end procedure

procedure about_box()
    sequence p1, p2
    p1 = zones[1][1]+{60, 5}
    p2 = zones[1][2]+{-60, 100}
    open_window(p1, p2)
    set_mouse_cursor(default_mouse_cursor)
    text_box(p1, p2, 260 &"PALETTE MANAGEMENT DEMO "& 256 &" by Colin Taylor"&
	"\nversion 1.0  -  12/23/98\n\nwritten in the EUPHORIA programming l"&
	"anguage\n-mouse pointer software by Pete Eberlein\n-graphic text so"&
	"ftware by Jiri Babor")
    exit_loop()
    close_window()
end procedure  -- about_box

procedure help()
    open_window({100,100}, {539, 379})
    box_3d(-1, 0, {120, 140}, {519, 339})
    define_button({449, 349}, {519, 369}, "Exit", 0, 1)
    text({130, 115}, {260, -8} &"OBLIGATORY HELP SCREEN")
    select_font(saf)
    text({130, 155}, "C'mon, you don't really need help, do you?  Just poke "&
	"around and see how \nit works.  A few little suggestions though... "&
	"\n\n     - notice how dithering can make a big improvement (in most"&
	" cases).\n\n     - turn off the system palette options and see if i"&
	"t affects the display.\n\n     - load your own image file.  Files b"&
	"igger than 320x240 pixels will be\n        trimmed (the view, that "&
	"is, not the file itself).\n\n     - turn of the auto file load and "&
	"auto refresh options.\n\nNow get on with it!")
    select_font(sbf)
    input_loop()
    close_window()
end procedure  -- help

procedure field(sequence msg)
-- clears field and displays msg in field
    select_font(saf)
    mouse_pointer(0)
    box(colors[2], 1, -1, zones[2][1], zones[2][2])
    text({125, 309}, msg)
    mouse_pointer(1)
    select_font(sbf)
end procedure  -- field

procedure display_palette(sequence pal, sequence xy, 
	    integer w, integer h, sequence colors)
-- colors = {foreground, background, highlight, shadow}
    sequence p1, p2
    if length(pal) then
	all_palette(pal/4)
    end if
    for j = 1 to 16 do
	for i = 1 to 16 do
	    p1 = {xy[1]+(i-1)*w+1, xy[2]+(j-1)*h+1}
	    p2 = {xy[1]+i*w-1, xy[2]+j*h-1}
	    if i+(j-1)*16 > length(pal) then
		box(colors[2], 1, 0, p1, p2)
		box_3d(-1, 0, p1, p2)
	    else
		box((i-1)+(j-1)*16, 1, 0, p1, p2)
	    end if
	end for
    end for
end procedure  -- display_palette

procedure refresh()
-- updates display
    sequence pal, new_image
    set_mouse_cursor(wait)
    pal = palettes[curr_pal]
    if dither then
	field("dithering image...")
	new_image = dither_image(pal, image)
    else
	new_image = convert_image(pal, image)
    end if
    if exact then
	new_image = systemize_image(palettes[5], new_image, exact-12)
    end if
    field(file_name)
    box(colors[2], 1, 0, {40, 50}, {40, 50}+max_size-1)
    box_3d(-1, 0, {40, 50}, {40, 50}+max_size-1)
    display_palette(new_image[1], {405, 92}, 11, 10, colors)
    display_image({40, 50}, new_image[2])
    set_mouse_cursor(IDC_ARROW)
end procedure  -- refresh

procedure auto_refresh()
    if auto then
	refresh()
    end if
end procedure  -- auto_refresh

procedure select_palette()
    if mbox != curr_pal then
	bullet(curr_pal, 1)
	curr_pal = mbox
	bullet(curr_pal, 2)
	auto_refresh()
    end if
end procedure  -- select_palette

procedure select_auto_load()
    load = not load
    bullet(10, load+1)
    button(1, not load)
end procedure  -- select_auto_load

procedure select_auto_refresh()
    auto = not auto
    bullet(11, auto+1)
    button(2, not auto)
    auto_refresh()
end procedure  -- select_auto_refresh

procedure select_dither()
    dither = not dither
    bullet(12, dither+1)
    auto_refresh()
end procedure  -- select_dither()

procedure select_exact()
    if mbox = exact then
	bullet(exact, 1)
	exact = 0
    elsif exact = 0 then
	exact = mbox
	bullet(exact, 2)
    else
	bullet(exact, 1)
	exact = mbox
	bullet(exact, 2)
    end if
    auto_refresh()
end procedure  -- select_exact

procedure load_file()
    object result
    result = read_bitmap(file_name)
    if atom(result) then
	field(file_err[result])
	exit_loop()
	field(file_name)
    else
	image = result
	if length(image[2]) > max_size[2] then
	    field("oversize image cropped")
	    image[2] = image[2][1..max_size[2]]
	end if
	if length(image[2][1]) > max_size[1] then
	    field("oversize image cropped")
	    for i = 1 to length(image[2]) do
		image[2][i] = image[2][i][1..max_size[1]]
	    end for
	end if
	palettes[1] = image[1]
	curr_pal = 1
	bullet(1, 2)
	for i = 2 to 9 do
	    if bullets[i][3] = SET then
		bullet(i, NORMAL)
		exit
	    end if
	end for
	if dither then
	    dither = 0
	    bullet(12, 1)
	end if
	refresh()
    end if
    button(1, 0)
end procedure  -- load_file

procedure edit_file_name()
    sequence result
    select_font(saf)
    setx(125)
    sety(309)
    mouse_pointer(0)
    result = prompt(file_name, 225)
    if result[1] = 27 then
	field(file_name)
    else
	file_name = result[2]
    end if
    mouse_pointer(1)
    select_font(sbf)
    if result[1] != 27 and load then
	load_file()
    else
	button(1, 1)
    end if
end procedure  -- edit_file_name

if graphics_mode(257) then
    puts(1, "Need Super VGA\n")
    abort(1)
end if  
vc = video_config()
saf = load_font("sysa13.f")
sbf = load_font("sysb13.f")
aus = load_font("austin36.f")
upc = load_font("upcase5.f")
select_font(aus)
colors = get_sys_colors()
box(colors[2], 1, 0, {0, 0}, vc[5..6]-1)  -- clear screen
set_background(CLEAR)
text({70, 5}, {-4, 257,'P',258,'a',259,'l',260,'e',261,'t',262,'t',257,'e',
    258,' ','M',259,'a',260,'n',261,'a',262,'g',257,'e',258,'m',259,'e',
    260,'n',261,'t',262,' ','D',257,'e',258,'m',259,'o'})
set_background(colors[2])
box_3d(-1, 0, {40, 50}, {40, 50}+max_size-1)  -- image area
-- set up palette area
box_3d(1, 0, {378, 50}, {600, 445})
select_font(sbf)
text({393, 60}, {260, -8} & "Current Palette:")
box_3d(1, 10, {402, 89}, {575, 247})
display_palette({}, {405, 92}, 11, 10, colors)
hex = "0123456789ABCDEF"
select_font(upc)
for i = 1 to 16 do
    text({405+i*11-7, 86}, {hex[i]})
    text({399, 92+i*10-7}, {hex[i]})
end for
select_font(saf)
id = routine_id("select_palette")
define_bullet({420, 270}, "Original Palette", id, 0)
define_bullet({420, 290}, "Win Palette (256 colors)", id, 0)
define_bullet({420, 310}, "Mac Palette (256 colors)", id, 0)
define_bullet({420, 330}, "Web Palette (216 colors)", id, 0)
define_bullet({420, 350}, "Dos Palette (16 colors)", id, 0)
define_bullet({420, 370}, "Gray Palette (16 colors)", id, 0)
define_bullet({420, 390}, "R/G/B Palette (5 colors)", id, 0)
define_bullet({420, 410}, "C/Y/M/K Palette (5 colors)", id, 0)
define_bullet({420, 430}, "B/W Palette (2 colors)", id, 0)
-- set up file name area
select_font(sbf)
text({50, 308}, {260, -8} & "File Name:") 
box_3d(-1, 0, {120, 305}, {360, 325})
-- set up options area
box_3d(1, 0, {40, 340}, {360, 412})
text({50, 347}, {260, -8} & "Options:") 
load = 1
auto = 1
dither = 0
exact = 14
select_font(saf)
define_bullet({60, 375}, "Auto File Load", routine_id("select_auto_load"), load+1)
define_bullet({60, 395}, "Auto Refresh", routine_id("select_auto_refresh"), auto+1)
define_bullet({175, 355}, "Dither Image (diffusion method)", routine_id("select_dither"), dither+1)
id = routine_id("select_exact")
define_bullet({175, 375}, "System Palette (nearest colors)", id, (exact = 13)+1)
define_bullet({175, 395}, "System Palette (exact colors)", id, (exact = 14)+1)
-- set up command area
select_font(sbf)
define_button({40, 425}, {111, 445}, "Load File", routine_id("load_file"), 1)
define_button({124, 425}, {194, 445}, "Refresh", routine_id("refresh"), 0)
define_button({207, 425}, {277, 445}, "Help", routine_id("help"), 1)
define_button({290, 425}, {360, 445}, "Exit", 0, 1)
define_zone({66,5}, {565, 39}, hand, routine_id("about_box"), 1)
define_zone({120, 305}, {360, 325}, ibeam, routine_id("edit_file_name"), 1)
-- load palettes
palettes = {{}}
add_palette(load_palette("win256.pal"))
add_palette(load_palette("mac256.pal"))
add_palette(load_palette("web.pal"))
add_palette(load_palette("dos16.pal"))
add_palette(load_palette("gray16.pal"))
add_palette({{0,0,0},{255,0,0},{0,255,0},{0,0,255},{255,255,255}})  -- RGB
add_palette({{0,0,0},{0,255,255},{255,255,0},{255,0,255},{255,255,255}})-- CYMK
add_palette({{0,0,0},{255,255,255}})  -- BW
-- load sample image
file_name = "sample.bmp"
field(file_name)
load_file()
-- main program loop
mouse_pointer(1)
input_loop()
if graphics_mode(-1) then
end if
